"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createCompressedBlobDefinition = exports.createCppJsStringDefinition = exports.npm = exports.objhash = exports.copyRecursive = exports.spawnBuildCommand = exports.pipeline = void 0;
const fs_1 = require("fs");
const crypto_1 = __importDefault(require("crypto"));
const child_process_1 = __importDefault(require("child_process"));
const util_1 = require("util");
const tar_1 = __importDefault(require("tar"));
const stream_1 = __importDefault(require("stream"));
const zlib_1 = __importDefault(require("zlib"));
const events_1 = require("events");
exports.pipeline = (0, util_1.promisify)(stream_1.default.pipeline);
async function spawnBuildCommand(command, options) {
    var _a, _b;
    options.logger.stepStarting(`Running ${command.join(' ')}`);
    const env = options.env;
    if (process.platform === 'win32') {
        const PATH = (_b = (_a = env.PATH) !== null && _a !== void 0 ? _a : env.Path) !== null && _b !== void 0 ? _b : env.path;
        delete env.PATH;
        delete env.Path;
        delete env.path;
        env.PATH = PATH;
    }
    const proc = child_process_1.default.spawn(command[0], command.slice(1), {
        stdio: 'inherit',
        ...options,
        env
    });
    const [code] = await (0, events_1.once)(proc, 'exit');
    if (code !== 0) {
        throw new Error(`Command failed: ${command.join(' ')} (code ${code})`);
    }
    options.logger.stepCompleted();
}
exports.spawnBuildCommand = spawnBuildCommand;
async function copyRecursive(sourceDir, targetDir) {
    await fs_1.promises.mkdir(targetDir, { recursive: true });
    await (0, exports.pipeline)(tar_1.default.c({
        cwd: sourceDir,
        gzip: false
    }, ['./']), tar_1.default.x({
        cwd: targetDir
    }));
}
exports.copyRecursive = copyRecursive;
function objhash(value) {
    return crypto_1.default.createHash('sha256')
        .update(JSON.stringify(value))
        .digest('hex')
        .slice(0, 32);
}
exports.objhash = objhash;
function npm() {
    if (process.env.npm_execpath) {
        return [process.execPath, process.env.npm_execpath];
    }
    else {
        return ['npm'];
    }
}
exports.npm = npm;
function createCppJsStringDefinition(fnName, source) {
    if (!source.length) {
        return `Local<String> ${fnName}(Isolate* isolate) { return String::Empty(isolate); }`;
    }
    const sourceAsCharCodeArray = new Uint16Array(source.length);
    let isAllLatin1 = true;
    for (let i = 0; i < source.length; i++) {
        const charCode = source.charCodeAt(i);
        sourceAsCharCodeArray[i] = charCode;
        isAllLatin1 && (isAllLatin1 = charCode <= 0xFF);
    }
    return `
  static const ${isAllLatin1 ? 'uint8_t' : 'uint16_t'} ${fnName}_source_[] = {
    ${sourceAsCharCodeArray}
  };
  static_assert(
    ${sourceAsCharCodeArray.length} <= v8::String::kMaxLength,
    "main script source exceeds max string length");
  Local<String> ${fnName}(Isolate* isolate) {
    return v8::String::NewFrom${isAllLatin1 ? 'One' : 'Two'}Byte(
      isolate,
      ${fnName}_source_,
      v8::NewStringType::kNormal,
      ${sourceAsCharCodeArray.length}).ToLocalChecked();
  }
  `;
}
exports.createCppJsStringDefinition = createCppJsStringDefinition;
async function createCompressedBlobDefinition(fnName, source) {
    const compressed = await (0, util_1.promisify)(zlib_1.default.brotliCompress)(source, {
        params: {
            [zlib_1.default.constants.BROTLI_PARAM_QUALITY]: zlib_1.default.constants.BROTLI_MAX_QUALITY,
            [zlib_1.default.constants.BROTLI_PARAM_SIZE_HINT]: source.length
        }
    });
    return `
  static const uint8_t ${fnName}_source_[] = {
    ${Uint8Array.prototype.toString.call(compressed) || '0'}
  };

#if __cplusplus >= 201703L
  [[maybe_unused]]
#endif
  static void ${fnName}_Read(char* dst) {
    size_t decoded_size = ${source.length};
    const auto result = BrotliDecoderDecompress(
      ${compressed.length},
      ${fnName}_source_,
      &decoded_size,
      reinterpret_cast<uint8_t*>(&dst[0]));
    assert(result == BROTLI_DECODER_RESULT_SUCCESS);
    assert(decoded_size == ${source.length});
  }

  std::string ${fnName}() {
    ${source.length === 0 ? 'return {};' : `
    std::string dst(${source.length}, 0);
    ${fnName}_Read(&dst[0]);
    return dst;`}
  }

  std::vector<char> ${fnName}Vector() {
    ${source.length === 0 ? 'return {};' : `
    std::vector<char> dst(${source.length});
    ${fnName}_Read(&dst[0]);
    return dst;`}
  }

  std::shared_ptr<v8::BackingStore> ${fnName}BackingStore() {
    std::string* str = new std::string(std::move(${fnName}()));
    return v8::SharedArrayBuffer::NewBackingStore(
      &str->front(),
      str->size(),
      [](void*, size_t, void* deleter_data) {
        delete static_cast<std::string*>(deleter_data);
      },
      static_cast<void*>(str));
  }

  v8::Local<v8::Uint8Array> ${fnName}Buffer(v8::Isolate* isolate) {
    ${source.length === 0 ? `
    auto array_buffer = v8::SharedArrayBuffer::New(isolate, 0);
    ` : `
    auto array_buffer = v8::SharedArrayBuffer::New(isolate, ${fnName}BackingStore());
    `}
    return v8::Uint8Array::New(array_buffer, 0, array_buffer->ByteLength());
  }
  `;
}
exports.createCompressedBlobDefinition = createCompressedBlobDefinition;
//# sourceMappingURL=helpers.js.map